;
; Full Speed USB Driver Subroutines Based On Microchip's Supplied C Code Version 1.2.
; Implementing a CDC Interface
;
; From usb9.c
; ************************************************************************************
; * Function:        void USBCheckStdRequest(void)                                   *
; *                                                                                  *
; * PreCondition:    None                                                            *
; * Input:           None                                                            *
; * Output:          None                                                            *
; * Side Effects:    None                                                            *
; * Overview:        check the setup data packet to see if it knows how to handle it *
; * Note:            None                                                            *
; ************************************************************************************
USBCheckStdRequest
    movlb high(_USB#RAM_START)
    Movf __SetupPkt,W
    Andlw 0x60
    Sublw (__STANDARD << 5)
    Bnz USBCheckStdRequestExit
    Movlw __SET_ADR
    Cpfseq SetupPkt_bRequest
    Bra USBCheckStdRequest1
    Movlw __ADR_PENDING_STATE
    Movwf __USB_DEVICE_STATE
    Bra Std@SetSessionOwnerUSB9
USBCheckStdRequest1
    Movlw __GET_DSC
    Cpfseq SetupPkt_bRequest
    Bra USBCheckStdRequest2
    Bra USBStdGetDscHandler
USBCheckStdRequest2
    Movlw __GET_CFG
    Cpfseq SetupPkt_bRequest
    Bra USBCheckStdRequest3

    Movlw (__USB_ACTIVE_CFG & 255)
    Movwf __Psrc
    Movlw ((__USB_ACTIVE_CFG >> 8) & 255)
    Movwf __PsrcH

    Movlw 1
    Movwf __wCount
    Clrf __wCountH
    Bcf __USB_STAT,__CTRL_TRF_MEM
    Bra Std@SetSessionOwnerUSB9
USBCheckStdRequest3
    Movlw __SET_CFG
    Cpfseq SetupPkt_bRequest
    Bra $ + 4
    Bra USBStdSetCfgHandler
    Movf SetupPkt_bRequest,W
    Bz USBStdGetStatusHandler
    Movlw __CLR_FEATURE
    Cpfseq SetupPkt_bRequest
    Bra $ + 4
    Bra USBStdFeatureReqHandler
    Movlw __SET_FEATURE
    Cpfseq SetupPkt_bRequest
    Bra $ + 4
    Bra USBStdFeatureReqHandler
    Movlw __GET_INTF
    Cpfseq SetupPkt_bRequest
    Bra USBCheckStdRequest8

    Movlw (__USB_ALT_INTF & 255)
    Movwf __Psrc
    Movlw ((__USB_ALT_INTF >> 8) & 255)
    Movwf __PsrcH

    Movf SetupPkt_bIntfID,W
    Addwf __Psrc,F
    Movlw 1
    Movwf __wCount
    Clrf __wCountH
    Bcf __USB_STAT,__CTRL_TRF_MEM
    Bra Std@SetSessionOwnerUSB9
USBCheckStdRequest8
    Movlw __SET_INTF
    Cpfseq SetupPkt_bRequest
    Return
    Lfsr 2,__USB_ALT_INTF
    Movf SetupPkt_bIntfID,W
    Movff SetupPkt_bAltID,PLUSW2
Std@SetSessionOwnerUSB9
    Movlw __MUID_USB9
    Movwf __CTRL_TRF_SESSION_OWNER
USBCheckStdRequestExit
    Return
; From usb9.c
; ************************************************************************************
; * Function:        void USBStdGetDscHandler(void)                                  *
; * PreCondition:    None                                                            *
; * Input:           None                                                            *
; * Output:          None                                                            *
; * Side Effects:    None                                                            *
; * Overview:        Handle the standard GET_DESCRIPTOR request.                     *
; *                  It utilises tables dynamically looks up descriptor size.        *
; *                  This routine should never have to be modified if the tables     *
; *                  in usbdsc.c are declared correctly.                             *
; * Note:            None                                                            *
; ************************************************************************************
USBStdGetDscHandler
    movlb high(_USB#RAM_START)
    Movlw 128
    Cpfseq SetupPkt_bmRequestType
    Return
    Movlw __DSC_DEV
    Cpfseq SetupPkt_bDscType
    Bra USBStdGetDscHandler1
NOLIST
if((__DeviceDescriptor & 255) == 0)
LIST
    Clrf __Psrc
    Clrf TBLPTRL
NOLIST
else
LIST
NOLIST
if((__DeviceDescriptor & 255) == 255)
LIST
    Setf __Psrc
    Setf TBLPTRL
NOLIST
else
LIST
    Movlw (__DeviceDescriptor & 255)
    Movwf __Psrc
    Movwf TBLPTRL
NOLIST
endif
NOLIST
endif
NOLIST
if(((__DeviceDescriptor >> 8) & 255) == 0)
LIST
    Clrf __PsrcH
    Clrf TBLPTRH
NOLIST
else
NOLIST
if(((__DeviceDescriptor >> 8) & 255) == 255)
LIST
    Setf __PsrcH
    Setf TBLPTRH
NOLIST
else
LIST
    Movlw ((__DeviceDescriptor >> 8) & 255)
    Movwf __PsrcH
    MOvwf TBLPTRH
NOLIST
endif
NOLIST
endif
LIST
    Tblrd*
    Movff TABLAT,__wCount
    Clrf __wCountH
    Bsf __USB_STAT,__CTRL_TRF_MEM
    Bra Std@SetSessionOwnerUSB9
USBStdGetDscHandler1
    Movlw __DSC_CFG
    Cpfseq SetupPkt_bDscType
    Bra USBStdGetDscHandler2
NOLIST
if((___Config1 & 255) == 0)
LIST
    Clrf __Psrc
NOLIST
else
LIST
NOLIST
if((___Config1 & 255) == 255)
LIST
    Setf __Psrc
NOLIST
else
LIST
    Movlw (___Config1 & 255)
    Movwf __Psrc
NOLIST
endif
NOLIST
endif
NOLIST
if(((___Config1 >> 8) & 255) == 0)
LIST
    Clrf __PsrcH
NOLIST
else
NOLIST
if(((___Config1 >> 8) & 255) == 255)
LIST
    Setf __PsrcH
NOLIST
else
LIST
    Movlw ((___Config1 >> 8) & 255)
    Movwf __PsrcH
NOLIST
endif
NOLIST
endif
LIST
NOLIST
if((__Config1Len & 255) == 0)
LIST
    Clrf TBLPTRL
NOLIST
else
LIST
NOLIST
if((__Config1Len & 255) == 255)
LIST
    Setf TBLPTRL
NOLIST
else
LIST
    Movlw (__Config1Len & 255)
    Movwf TBLPTRL
NOLIST
endif
NOLIST
endif
NOLIST
if(((__Config1Len >> 8) & 255) == 0)
LIST
    Clrf TBLPTRH
NOLIST
else
NOLIST
if(((__Config1Len >> 8) & 255) == 255)
LIST
    Setf TBLPTRH
NOLIST
else
LIST
    Movlw ((__Config1Len >> 8) & 255)
    Movwf TBLPTRH
NOLIST
endif
NOLIST
endif
LIST
    Tblrd*+
    Movff TABLAT,__wCount
    Tblrd*+
    Tblrd*
    Movff TABLAT,__wCountH
    Bsf __USB_STAT,__CTRL_TRF_MEM
    Bra Std@SetSessionOwnerUSB9
USBStdGetDscHandler2
    Movlw __DSC_STR
    Cpfseq SetupPkt_bDscType
    Return
    Clrf TBLPTRH
    Rlncf SetupPkt_bDscIndex,W
    Addlw (__USB_SD_Ptr & 255)
    Movwf TBLPTRL
    Movlw ((__USB_SD_Ptr >> 8) & 255)
    addwfc TBLPTRH,F
    Tblrd*+
    Movff TABLAT,__Psrc
    Tblrd*
    Movf TABLAT,W
    Movwf __PsrcH
    Movwf TBLPTRH
    Movff __Psrc,TBLPTRL
    Tblrd*
    Movff TABLAT,__wCount
    Clrf __wCountH
    Bsf __USB_STAT,__CTRL_TRF_MEM
    Bra Std@SetSessionOwnerUSB9
; From usb9.c
; ************************************************************************************
; * Function:        void USBStdSetCfgHandler(void)                                  *
; * PreCondition:    None                                                            *
; * Input:           None                                                            *
; * Output:          None                                                            *
; * Side Effects:    None                                                            *
; * Overview:        First disable all endpoints by clearing UEP registers           *
; *                  Then initialises endpoints specified in the modifiable section. *
; * Note:            None                                                            *
; ************************************************************************************
USBStdSetCfgHandler
    movlb high(_USB#RAM_START)
    Movlw __MUID_USB9
    Movwf __CTRL_TRF_SESSION_OWNER
    Lfsr 2,UEP1
NOLIST
#if (__Ep0bO == 512) ; Smaller USB devices only have 8 end points
LIST
    Movlw 7
NOLIST
#else
LIST
    Movlw 15
NOLIST
#endif
LIST
    Clrf POSTINC2
    Decfsz WREG,F
    Bra $ - 4
    Lfsr 2,__USB_ALT_INTF
    Movlw __MAX_NUM_INT
    Clrf POSTINC2
    Decfsz WREG,F
    Bra $ - 4
    Movf SetupPkt_bCfgValue,W
    Movwf __USB_ACTIVE_CFG
    Bnz USBStdSetCfgHandler1
    Movlw __ADDRESS_STATE
    Movwf __USB_DEVICE_STATE
    Return
USBStdSetCfgHandler1
    Movlw __CONFIGURED_STATE
    Movwf __USB_DEVICE_STATE
    Bra CDCInit@EP
; From usb9.c
; ************************************************************************************
; * Function:        void USBStdGetStatusHandler(void)                               *
; * PreCondition:    None                                                            *
; * Input:           None                                                            *
; * Output:          None                                                            *
; * Side Effects:    None                                                            *
; * Overview:        Handle the standard GET_STATUS request                          *
; * Note:            None                                                            *
; ************************************************************************************
USBStdGetStatusHandler
    movlb high(_USB#RAM_START)
    Clrf __CtrlTRFdata#0
    Clrf __CtrlTRFdata#1
    Movf __SetupPkt,W
    Andlw 0x1F
    Sublw __RCPT_DEV
    Bnz USBStdGetStatusHandler1
NOLIST
ifdef USB_BUS_SELF_POWER#REQ
LIST
    Bsf __CtrlTRFdata,0
NOLIST
endif
LIST
    Btfsc __USB_STAT,__REMOTEWAKEUP
    Bsf __CtrlTRFdata,1
    Bra Std@GetStatusSetSessionOwner
USBStdGetStatusHandler1
    Movf __SetupPkt,W
    Andlw 0x1F
    Sublw __RCPT_INTF
    Bnz USBStdGetStatusHandler2
    Bra Std@GetStatusSetSessionOwner
USBStdGetStatusHandler2
    Movf __SetupPkt,W
    Andlw 0x1F
    Sublw __RCPT_EP
    Bnz USBStdGetStatusHandler3
    Rcall Calc@EPAddress
    Movf INDF2,W
    Andlw _BSTALL
    Bz Std@GetStatusSetSessionOwner
    Bsf __CtrlTRFdata,0
Std@GetStatusSetSessionOwner
    Movlw __MUID_USB9
    Movwf __CTRL_TRF_SESSION_OWNER
USBStdGetStatusHandler3
    Movlw __MUID_USB9
    Cpfseq __CTRL_TRF_SESSION_OWNER
    Return

    Movlw (__CtrlTRFdata & 255)
    Movwf __Psrc
    Movlw ((__CtrlTRFdata >> 8) & 255)
    Movwf __PsrcH

    Movlw 2
    Movwf __wCount
    Clrf __wCountH
    Bcf __USB_STAT,__CTRL_TRF_MEM
    Return
; From usb9.c
; ************************************************************************************
; * Function:        void USBStdFeatureReqHandler(void)                              *
; * PreCondition:    None                                                            *
; * Input:           None                                                            *
; * Output:          None                                                            *
; * Side Effects:    None                                                            *
; * Overview:        handle the standard Set & Clear Features requests               *
; * Note:            None                                                            *
; ************************************************************************************
USBStdFeatureReqHandler
    movlb high(_USB#RAM_START)
    Movlw __DEVICE_REMOTE_WAKEUP
    Cpfseq SetupPkt_bFeature
    Bra USBStdFeatureReqHandler1
    Movf __SetupPkt,W
    Andlw 0x1F
    Sublw __RCPT_DEV
    Bnz USBStdFeatureReqHandler1
    Bsf __USB_STAT,__REMOTEWAKEUP
    Movlw __SET_FEATURE
    Cpfseq SetupPkt_bRequest
    Bcf __USB_STAT,__REMOTEWAKEUP
    Bra Std@SetSessionOwnerUSB9
USBStdFeatureReqHandler1
    Movlw __ENDPOINT_HALT
    Cpfseq SetupPkt_bFeature
USBStdFeatureReqHandlerExit
    Return
    Movf __SetupPkt,W
    Andlw 0x1F
    Sublw __RCPT_EP
    Bnz USBStdFeatureReqHandlerExit
    Movf SetupPkt_bEPID,W
    Andlw 0x0F
    Bz USBStdFeatureReqHandlerExit
    Rcall Calc@EPAddress
    Movlw __SET_FEATURE
    Cpfseq SetupPkt_bRequest
    Bra USBStdFeatureReqHandler2
    Movlw _USIE | _BSTALL
    Movwf INDF2
    Bra Std@SetSessionOwnerUSB9
USBStdFeatureReqHandler2
    movlw _UCPU | _DAT1
    Btfss SetupPkt_bEPID,__tEPDIR
    Movlw _USIE | _DAT0 | _DTSEN
    Movwf INDF2
    Bra Std@SetSessionOwnerUSB9
Calc@EPAddress
    Lfsr 2,__Ep0bO
    Rlncf SetupPkt_bEPID,W
    Rlcf SetupPkt_bEPID,W
    Rlncf WREG,F
    Rlncf WREG,F
    Addwf FSR2L,F
    Return
; From usbctrltrf.c
; ************************************************************************************
; * Function:        void USBCtrlEPService(void)                                     *
; * PreCondition:    USTAT is loaded with a valid endpoint address.                  *
; * Input:           None                                                            *
; * Output:          trnifcleared - 0xFF for TRNIF cleared, for not                  *
; * Side Effects:    None                                                            *
; * Overview:        USBCtrlEPService checks for three transaction types that        *
; *                  it knows how to service and services them:                      *
; *                  1. EP0 SETUP                                                    *
; *                  2. EP0 OUT                                                      *
; *                  3. EP0 IN                                                       *
; *                  It ignores all other types (i.e. EP1, EP2, etc.)                *
; * Note:            None                                                            *
; ************************************************************************************
USBCtrlEPService
    Clrf __TRNIFCLEARED
    Movlw __EP00_OUT
    Cpfseq USTAT
    Bra USBCtrlEPService1
    Bcf UIR,PP_TRNIF
    Setf __TRNIFCLEARED
    Movf Ep0bO_Stat,W
    Andlw 0x3C
    Sublw (__SETUP_TOKEN << 2)
    Bz USBCtrlTrfSetupHandler
    Bra USBCtrlTrfOutHandler
USBCtrlEPService1
    Movlw __EP00_IN
    Cpfseq USTAT
    Return
    Bcf UIR,PP_TRNIF
    Setf __TRNIFCLEARED
; From usbctrltrf.c
; ************************************************************************************
; * Function:        void USBCtrlTrfInHandler(void)                                  *
; * PreCondition:    None                                                            *
; * Input:           None                                                            *
; * Output:          None                                                            *
; * Side Effects:    None                                                            *
; * Overview:        Handle an IN transaction according to                           *
; *                  which control transfer state is currently active.               *
; *                                                                                  *
; * Note:            A Set Address Request must not change the acutal address        *
; *                  of the device until the completion of the control               *
; *                  transfer. The end of the control transfer for Set Address       *
; *                  Request is an IN transaction. Therefore it is necessary         *
; *                  to service this unique situation when the condition is          *
; *                  right. Macro mUSBCheckAdrPendingState is defined in             *
; *                  usb9.h and its function is to specifically service this         *
; *                  event.                                                          *
; ************************************************************************************
USBCtrlTrfInHandler
    Movlw __ADR_PENDING_STATE
    Cpfseq __USB_DEVICE_STATE
    Bra USBCtrlTrfInHandler1
    Movf SetupPkt_bDevADR,W
NOLIST
if (UADDR < 0XF60)
LIST
    Movlb ((UADDR >> 8) & 255)
NOLIST
endif
LIST
    Movwf UADDR
    movlb high(_USB#RAM_START)
    Movlw __ADDRESS_STATE
    Btfsc STATUS,Z
    Movlw __DEFAULT_STATE
    Movwf __USB_DEVICE_STATE
USBCtrlTrfInHandler1
    Movlw __CTRL_TRF_TX
    Cpfseq __CTRL_TRF_STATE
    Bra USBPrepareForNextSetupTrf
    Rcall USBCtrlTrfTxService
    movlb high(_USB#RAM_START)
    Movlw __SHORT_PKT_SENT
    Cpfseq __SHORT_PKT_STATUS
    Bra    CTL@TRFINHDL2
    Movlw _USIE | _BSTALL
    Movwf Ep0bI_Stat
    Return
CTL@TRFINHDL2
    Movlw _USIE | _DAT1 | _DTSEN
    Btfsc Ep0bI_Stat,__DTS
    Movlw _USIE | _DAT0 | _DTSEN
    Movwf Ep0bI_Stat
    Return
; From usbctrltrf.c
; ************************************************************************************
; * Function:        void USBCtrlTrfSetupHandler(void)                               *
; * PreCondition:    SetupPkt buffer is loaded with valid USB Setup Data             *
; * Input:           None                                                            *
; * Output:          None                                                            *
; * Side Effects:    None                                                            *
; * Overview:        Task dispatcher that has 3 stages.                              *
; *                  1. Initialises the control transfer state machine.              *
; *                  2. Calls on each of the module that may know how to             *
; *                     service the Setup Request from the host.                     *
; *                     Module Example: USB9, HID, CDC, MSD, ...                     *
; *                     As new classes are added, ClassReqHandler table in           *
; *                     usbdsc.c should be updated to call all available             *
; *                     class handlers.                                              *
; *                  3. Once each of the modules has had a chance to check if        *
; *                     it is responsible for servicing the request, stage 3         *
; *                     then checks direction of the transfer to determine how       *
; *                     to prepare EP0 for the control transfer.                     *
; *                     Refer to USBCtrlEPServiceComplete() for more details.        *
; * Note:            USB Firmware has three different states for                     *
; *                  the control transfer state machine:                             *
; *                  1. WAIT_SETUP                                                   *
; *                  2. CTRL_TRF_TX                                                  *
; *                  3. CTRL_TRF_RX                                                  *
; *                  Refer to firmware manual to find out how one state              *
; *                  is transitioned to another.                                     *
; *                                                                                  *
; *                  A Control Transfer is composed of many USB transactions.        *
; *                  When transferring data over multiple transactions,              *
; *                  it is important to keep track of data source, data              *
; *                  destination, and data count. These three parameters are         *
; *                  stored in pSrc,pDst, and wCount. A flag is used to              *
; *                  note if the data source is from ROM or RAM.                     *
; ************************************************************************************
USBCtrlTrfSetupHandler
    Movlw _UCPU
    Btfsc Ep0bI_Stat,__UOWN
    Movwf Ep0bI_Stat
    Movlw __SHORT_PKT_NOT_USED
    Movwf __SHORT_PKT_STATUS
    Movlw __WAIT_SETUP
    Movwf __CTRL_TRF_STATE
    Clrf __CTRL_TRF_SESSION_OWNER
    Clrf __wCount
    Clrf __wCountH
    Rcall USBCheckStdRequest
    Movf __CTRL_TRF_SESSION_OWNER,W
    Bnz USBCtrlEPServiceComplete
    Rcall Check@CDCRequest
    Bra USBCtrlEPServiceComplete
; From usbctrltrf.c
; ************************************************************************************
; * Function:        void USBCtrlTrfOutHandler(void)                                 *
; * PreCondition:    None                                                            *
; * Input:           None                                                            *
; * Output:          None                                                            *
; * Side Effects:    None                                                            *
; * Overview:        Handle an OUT transaction according to                          *
; *                  which control transfer state is currently active.               *
; * Note:            Note that if the the control transfer was from                  *
; *                  host to device, the session owner should be notified            *
; *                  at the end of each OUT transaction to service the               *
; *                  received data.                                                  *
; ************************************************************************************
USBCtrlTrfOutHandler
    Movlw __CTRL_TRF_RX
    Cpfseq __CTRL_TRF_STATE
    Bra USBPrepareForNextSetupTrf
    Rcall USBCtrlTrfRxService
    Movlw _USIE | _DAT1 | _DTSEN
    Btfsc Ep0bO_Stat,__DTS
    Movlw _USIE | _DAT0 | _DTSEN
    Movwf Ep0bO_Stat
    Return
; From usbctrltrf.c
; ************************************************************************************
; * Function:        void USBCtrlTrfTxService(void)                                  *
; * PreCondition:    pSrc, wCount, and USB_STAT.__CTRL_TRF_MEM are setup properly.   *
; * Input:           None                                                            *
; * Output:          None                                                            *
; * Side Effects:    None                                                            *
; * Overview:        Routine should be called from only two places.                  *
; *                  One from USBCtrlEPServiceComplete() and one from                *
; *                  USBCtrlTrfInHandler(). It takes care of managing a              *
; *                  transfer over multiple USB transactions.                        *
; * Note:            Copies one packet-ful of data pSrc (either ROM or RAM) to       *
; *                  EP0 IN buffer. It then updates pSrc to be ready for next        *
; *                  piece.                                                          *
; *                  Routine works with isochronous endpoint larger than             *
; *                  256 bytes and is shown here as an example of how to deal        *
; *                  with BC9 and BC8. In reality, a control endpoint can never      *
; *                  be larger than 64 bytes.                                        *
; ************************************************************************************
USBCtrlTrfTxService
    movlb high(_USB#RAM_START)
    Movf __wCount,W
    Movwf __USB_TEMP_VAR
    Movf __wCountH,W
    Movwf __USB_TEMP_VARH
    Sublw ((__EP0_BUFF_SIZE >> 8) & 255)
    Bnz CTL@TRFTXSVCNE
    Movf __wCount,W
    Sublw (__EP0_BUFF_SIZE & 255)
    Bz CTL@TRFTXSVCSB
CTL@TRFTXSVCNE
    Bnc CTL@TRFTXSVCMX
    Movlw __SHORT_PKT_NOT_USED
    Cpfseq __SHORT_PKT_STATUS
    Bra CTL@TRFTXSVCNE1
    Movlw __SHORT_PKT_PENDING
    Movwf __SHORT_PKT_STATUS
    Bra CTL@TRFTXSVCSB
CTL@TRFTXSVCNE1
    Movlw __SHORT_PKT_PENDING
    Cpfseq __SHORT_PKT_STATUS
    Bra CTL@TRFTXSVCSB
    Movlw __SHORT_PKT_SENT
    Movwf __SHORT_PKT_STATUS
    Bra CTL@TRFTXSVCSB
CTL@TRFTXSVCMX
    Movlw __EP0_BUFF_SIZE
    Movwf __USB_TEMP_VAR
    Clrf __USB_TEMP_VARH
CTL@TRFTXSVCSB
    Movf __USB_TEMP_VAR,W
    Movwf Ep0bI_Cnt
    Subwf __wCount,F
    Movf __USB_TEMP_VARH,W
    Subwfb __wCountH,F
    Movf Ep0bI_Stat,W
    Andlw 0XFC
    Iorwf __USB_TEMP_VARH,W
    Movwf Ep0bI_Stat
    Lfsr 2,__CtrlTRFdata
    Movf __USB_TEMP_VARH,W
    Bnz CTL@TRFTXSVCROMRAM
    Movf __USB_TEMP_VAR,W
    Bz CTL@TRFTXSVCEXIT
CTL@TRFTXSVCROMRAM
    Btfss __USB_STAT,__CTRL_TRF_MEM
    Bra CTL@TRFTXSVCRM
    Movff __Psrc,TBLPTRL
    Movff __PsrcH,TBLPTRH
CTL@TRFTXSVCROMLP
    Tblrd*+
    Movff TABLAT,POSTINC2
    Tblrd*+
    Decf __USB_TEMP_VAR,F
    Bnz CTL@TRFTXSVCROMLP
    Decf __USB_TEMP_VARH,F
    Bc CTL@TRFTXSVCROMLP
    Movff TBLPTRL,__Psrc
    Movff TBLPTRH,__PsrcH
    Return
CTL@TRFTXSVCRM
    Movff __Psrc,FSR1L
    Movff __PsrcH,FSR1H
CTL@TRFTXSVCRMLP
    Movff POSTINC1,POSTINC2
    Decf __USB_TEMP_VAR,F
    Bnz CTL@TRFTXSVCRMLP
    Decf __USB_TEMP_VARH,F
    Bc CTL@TRFTXSVCRMLP
    Movff FSR1L,__Psrc
    Movff FSR1H,__PsrcH
CTL@TRFTXSVCEXIT
    Return
; From usbctrltrf.c
; ******************************************************************************
; * Function:        void USBCtrlTrfRxService(void)                            *
; * PreCondition:    pDst and wCount are setup properly.                       *
; *                  pSrc is always &CtrlTrfData                               *
; *                  __USB_STAT.__CTRL_TRF_MEM is always RAM.                  *
; *                  wCount should be set to 0 at the start of each control    *
; *                  transfer.                                                 *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        Transfers bytes received, at EP0 OUT buffer CtrlTrfData,  *
; *                  to user's buffer pointed by pDst.                         *
; *                  This routine only knows how to handle raw byte data.      *
; * Note:            None                                                      *
; ******************************************************************************
USBCtrlTrfRxService
    movlb high(_USB#RAM_START)
    Movf Ep0bO_Cnt,W
    Movwf __USB_TEMP_VAR
    Addwf __wCount,F
    Movf Ep0bO_Stat,W
    Andlw 3
    Movwf __USB_TEMP_VARH
    addwfc __wCountH,F
    Lfsr 1,__CtrlTRFdata
    Movff __Pdst,FSR2L
    Movff __PdstH,FSR2H
    Movf __USB_TEMP_VARH,W
    Bnz USBCtrlTrfRxServiceLoop
    Movf __USB_TEMP_VAR,W
    Bz USBCtrlTrfRxServiceExit
USBCtrlTrfRxServiceLoop
    Movff POSTINC1,POSTINC2
    Decf __USB_TEMP_VAR,F
    Bnz USBCtrlTrfRxServiceLoop
    Decf __USB_TEMP_VARH,F
    Bc USBCtrlTrfRxServiceLoop
    Movff FSR2L,__Pdst
    Movff FSR2H,__PdstH
USBCtrlTrfRxServiceExit
    Return
; From usbctrltrf.c
; ******************************************************************************
; * Function:        void USBCtrlEPServiceComplete(void)                       *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        wrap up the ramaining tasks in servicing a Setup Request  *
; *                  Its main task is to set the endpoint                      *
; *                  controls appropriately for a given situation. See code    *
; *                  below.                                                    *
; *                  There are three main scenarios:                           *
; *                  a) There was no handler for the Request, in this case     *
; *                     a STALL should be sent out.                            *
; *                  b) The host has requested a read control transfer,        *
; *                     endpoints are required to be setup in a specific way.  *
; *                  c) The host has requested a write control transfer, or    *
; *                     a control data stage is not required, endpoints are    *
; *                     required to be setup in a specific way.                *
; *                                                                            *
; *                  Packet processing is resumed by clearing PKTDIS bit.      *
; * Note:            None                                                      *
; ******************************************************************************
USBCtrlEPServiceComplete
    Bcf UCON,PP_PKTDIS
    Movf __CTRL_TRF_SESSION_OWNER
    Bnz USBCtrlEPServiceComplete1
    Movlw __EP0_BUFF_SIZE
    Movwf Ep0bO_Cnt
    Movlw (__SetupPkt & 255)
    Movwf Ep0bO_AdrL
    Movlw ((__SetupPkt >> 8) & 255)
    Movwf Ep0bO_AdrH
    Movlw _USIE | _BSTALL
    Movwf Ep0bO_Stat
    Movwf Ep0bI_Stat
    Bra USBCtrlEPServiceCompleteExit
USBCtrlEPServiceComplete1
    Btfss __SetupPkt,__tDATADIR
    Bra USBCtrlEPServiceComplete2
    Movf __wCountH,W
    Subwf SetupPkt_wLengthH,W
    Bnc USBCtrlEPServiceCompleteCopy
    Bnz USBCtrlEPServiceComplete11
    Movf __wCount,W
    Subwf SetupPkt_wLength,W
    Bc USBCtrlEPServiceComplete11
USBCtrlEPServiceCompleteCopy
    Movff SetupPkt_wLength,__wCount
    Movff SetupPkt_wLengthH,__wCountH
USBCtrlEPServiceComplete11
    Rcall USBCtrlTrfTxService
    Movlw __CTRL_TRF_TX
    Movwf __CTRL_TRF_STATE

    Movlw __EP0_BUFF_SIZE
    Movwf Ep0bO_Cnt
    Movlw (__SetupPkt & 255)
    Movwf Ep0bO_AdrL
    Movlw ((__SetupPkt >> 8) & 255)
    Movwf Ep0bO_AdrH
    Movlw _USIE
    Movwf Ep0bO_Stat
    Movlw (__CtrlTRFdata & 255)
    Movwf Ep0bI_AdrL
    Movlw ((__CtrlTRFdata >> 8) & 255)
    Movwf Ep0bI_AdrH
    Movlw _USIE | _DAT1 | _DTSEN
    Movwf Ep0bI_Stat
    Bra USBCtrlEPServiceCompleteExit

USBCtrlEPServiceComplete2
    Movlw __CTRL_TRF_RX
    Movwf __CTRL_TRF_STATE
    Clrf Ep0bI_Cnt
    Movlw _USIE | _DAT1 | _DTSEN
    Movwf Ep0bI_Stat
    Movlw __EP0_BUFF_SIZE
    Movwf Ep0bO_Cnt
    Movlw (__CtrlTRFdata & 255)
    Movwf Ep0bO_AdrL
    Movlw ((__CtrlTRFdata >> 8) & 255)
    Movwf Ep0bO_AdrH
    Movlw _USIE | _DAT1 | _DTSEN
    Movwf Ep0bO_Stat
USBCtrlEPServiceCompleteExit
    Bcf UCON,PP_PKTDIS
    Return
; From usbctrltrf.c
;/******************************************************************************
; * Function:        void USBPrepareForNextSetupTrf(void)                      *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        Forces EP0 OUT to be ready for a new Setup                *
; *                  transaction, and forces EP0 IN to be owned by CPU.        *
; * Note:            None                                                      *
; ******************************************************************************
USBPrepareForNextSetupTrf
    movlb high(_USB#RAM_START)
    Movlw __CTRL_TRF_RX
    Cpfseq __CTRL_TRF_STATE
    Bra PREP@FNSTRF1
    Btfss UCON,PP_PKTDIS
    Bra PREP@FNSTRF1
    Movlw __EP0_BUFF_SIZE
    Cpfseq Ep0bO_Cnt
    Bra PREP@FNSTRF1
    Movf Ep0bO_Stat,W
    Andlw 0X3C
    Sublw (__SETUP_TOKEN << 2)
    Bnz PREP@FNSTRF1
    Btfsc Ep0bI_Stat,__UOWN
    Bra PREP@FNSTRF1
    Lfsr 1,__CtrlTRFdata
    Lfsr 2,__SetupPkt
    Movlw __EP0_BUFF_SIZE
PREP@FNSTRFLP
    Movff POSTINC1,POSTINC2
    Decfsz WREG,F
    Bra PREP@FNSTRFLP
    Movlw (__SetupPkt & 255)
    Movwf Ep0bO_AdrL
    Movlw ((__SetupPkt >> 8) & 255)
    Movwf Ep0bO_AdrH
    Return
PREP@FNSTRF1
    Movlw __WAIT_SETUP
    Movwf __CTRL_TRF_STATE
    Movlw __EP0_BUFF_SIZE
    Movwf Ep0bO_Cnt
    Movlw (__SetupPkt & 255)
    Movwf Ep0bO_AdrL
    Movlw ((__SetupPkt >> 8) & 255)
    Movwf Ep0bO_AdrH
    Movlw _USIE | _DAT0 | _DTSEN | _BSTALL
    Movwf Ep0bO_Stat
    Movlw _UCPU
    Movwf Ep0bI_Stat
    Return
; From usbdrv.c
; ******************************************************************************
; * Function:        void InitialiseUSB@Driver(void)                           *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        Initialise variables used by the USB library routines     *
; * Note:            None                                                      *
; ******************************************************************************

InitialiseUSB@Driver
    Clrf TBLPTRU
NOLIST
if (UCFG < 0X0F60)
LIST
    Movlb ((UCFG >> 8) & 255)
NOLIST
endif
LIST
    Movlw __UCFG_VAL
    Movwf UCFG
    movlb high(_USB#RAM_START)
    Clrf __USB_DEVICE_STATE
    Clrf __USB_STAT
    Clrf __USB_ACTIVE_CFG
    Bra CDCINIT@EP
; From usbdrv.c
; ******************************************************************************
; * Function:        void USBCheckBusStatus(void)                                *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        Enable/Disable the USB module by monitoring               *
; *                  the USB power signal.                                     *
; * Note:            None                                                      *
; ******************************************************************************
Check@BusStatus
    Clrf TBLPTRU
    movlb high(_USB#RAM_START)
NOLIST
ifdef USB_BUS_SENSE_IO#REQ
LIST
    Btfss USB_BUS#SENSE
    Bra Check@BusStatusDetached
NOLIST
endif
LIST
    Btfss UCON,PP_USBEN
    Rcall MODULE@ENABLE
NOLIST
ifdef USB_BUS_SENSE_IO#REQ
LIST
    Bra CHK@BSTAT1
CHK@BSTATDETACHED
    Btfsc UCON,PP_USBEN
    Rcall MODULE@DISABLE
NOLIST
endif
LIST
CHK@BSTAT1
    Movlw __ATTACHED_STATE
    Cpfseq __USB_DEVICE_STATE
    Return
    Btfsc UCON,PP_SE0
    Return
    Clrf UIR
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Clrf UIE
    Bsf UIE,PP_URSTIE
    Bsf UIE,PP_IDLEIE
    movlb high(_USB#RAM_START)
    Movlw __POWERED_STATE
    Movwf __USB_DEVICE_STATE
    Return
; From usbdrv.c
; ******************************************************************************
; * Function:        void USBModuleEnable(void)                                *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        This routine enables the USB module.                      *
; *                  An end designer should never have to call this routine    *
; *                  manually. This routine should only be called from         *
; *                  USBCheckBusStatus().                                      *
; * Note:            See USBCheckBusStatus() for more information.             *
; ******************************************************************************
Module@Enable
    Clrf UCON
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Clrf UIE
    Bsf UCON,PP_USBEN
    movlb high(_USB#RAM_START)
    Movlw __ATTACHED_STATE
    Movwf __USB_DEVICE_STATE
    Return
; From usbdrv.c
; ******************************************************************************
; * Function:        void USBModuleDisable(void)                               *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        This routine disables the USB module.                     *
; *                  An end designer should never have to call this routine    *
; *                  manually. This routine should only be called from         *
; *                  USBCheckBusStatus().                                      *
; * Note:            See USBCheckBusStatus() for more information.             *
; ******************************************************************************
Module@Disable
    Clrf UCON
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Clrf UIE
    movlb high(_USB#RAM_START)
    Clrf __USB_DEVICE_STATE
    Return
; From usbdrv.c
; ******************************************************************************
; * Function:        void USBDriverService(void)                                 *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        Manage all USB interrupts.                                *
; * Note:            Device state transitions through the following stages:    *
; *                  DETACHED -> ATTACHED -> POWERED -> DEFAULT ->             *
; *                  ADDRESS_PENDING -> ADDRESSED -> CONFIGURED -> READY       *
; ******************************************************************************
Driver@Service
    movlb high(_USB#RAM_START)
    Movlw __DETACHED_STATE
    Subwf __USB_DEVICE_STATE,W
    Bz Driver@ServiceExit
    Btfss UIR,PP_ACTVIF
    Bra Driver@Service1
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Btfsc UIE,PP_ACTVIE
    Rcall Wake@FromSuspend
Driver@Service1
    Btfsc UCON,PP_SUSPND
    Return
    Btfss UIR,PP_URSTIF
    Bra $ + 6
    Btfsc UIE,PP_URSTIE
    Rcall Protocol@ResetHandler
    Btfss UIR,PP_IDLEIF
    Bra Driver@Service2
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Btfsc UIE,PP_IDLEIE
    Rcall Suspend@USB
Driver@Service2
    Btfss UIR,PP_SOFIF
    Bra Driver@Service3
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Btfsc UIE,PP_SOFIE
    Bcf UIR,PP_SOFIF
Driver@Service3
    Btfss UIR,PP_STALLIF
    Bra Driver@Service4
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Btfsc UIE,PP_STALLIE
    Rcall Stall@Handler
Driver@Service4
    Btfss UIR,PP_UERRIF
    Bra Driver@Service5
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Btfsc UIE,PP_UERRIE
    Bcf UIR,PP_UERRIF
Driver@Service5
    Movlw __DEFAULT_STATE
    Subwf __USB_DEVICE_STATE,W
    Bnc Driver@ServiceExit
    Btfss UIR,PP_TRNIF
    Bra Driver@ServiceExit
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Btfss UIE,PP_TRNIE
    Bra Driver@ServiceExit
    Rcall USBCtrlEPService
    Bcf UIR,PP_TRNIF
Driver@ServiceExit
    Return
; From usbdrv.c
; ******************************************************************************
; * Function:        void USBSuspend(void)                                    *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        None                                                      *
; * Note:            None                                                      *
; ******************************************************************************
Suspend@USB
; Note: Do not clear UIRbits.ACTVIF here!
; Reason:
; ACTVIF is only generated once an IDLEIF has been generated.
; This is a 1:1 ratio interrupt generation.
; For every IDLEIF, there will be only one ACTVIF regardless of
; the number of subsequent bus transitions.
;
; If the ACTIF is cleared here, a problem could occur when:
; [       IDLE       ][bus activity ->
; <--- 3 ms ----->     ^
;                ^     ACTVIF=1
;                IDLEIF=1
;  #           #           #           #   (#=Program polling flags)
;                          ^
;                          This polling loop will see both
;                          IDLEIF=1 and ACTVIF=1.
;                          However, the program services IDLEIF first
;                          because ACTIVIE=0.
;                          If this routine clears the only ACTIVIF,
;                          then it can never get out of the suspend
;                          mode.
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Bsf UIE,PP_ACTVIE
    Bcf UIR,PP_IDLEIF
    Bsf UCON,PP_SUSPND
    Return
; From usbdrv.c
; ******************************************************************************
; * Function:        void USBWakeFromSuspend(void)                               *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        None                                                      *
; * Note:            None                                                      *
; ******************************************************************************
Wake@FromSuspend
    Bcf UCON,PP_SUSPND
NOLIST
if (UIE < 0X0F60)
LIST
    Movlb ((UIE >> 8) & 255)
NOLIST
endif
LIST
    Bcf UIE, PP_ACTVIE
    Bcf UIR,PP_ACTVIF
    Btfsc UIR,PP_ACTVIF
    Bra $ - 4
    Return
; From usbdrv.c
; *******************************************************************************
; * Function:        void USBRemoteWakeup(void)                                   *
; * PreCondition:    None                                                       *
; * Input:           None                                                       *
; * Output:          None                                                       *
; * Side Effects:    None                                                       *
; * Overview:        Function should be called by user when the device          *
; *                  is waken up by an external stimulus other than ACTIVIF.    *
; *                  Please read the note below to understand the limitations.  *
; * Note:            The modifiable section in this routine should be changed   *
; *                  to meet the application needs. Current implementation      *
; *                  temporary blocks other functions from executing for a      *
; *                  period of 1-13 ms depending on the core frequency.         *
; *                                                                             *
; *                  According to USB 2.0 specification section 7.1.7.7,        *
; *                  "The remote wakeup device must hold the resume signaling   *
; *                  for at lest 1 ms but for no more than 15 ms."              *
; *                  The idea here is to use a delay counter loop, using a      *
; *                  common value that would work over a wide range of core     *
; *                  frequencies.                                               *
; *                  That value selected is 1800. See table below:              *
; *                  ========================================================== *
; *                  Core Freq(MHz)      MIP         RESUME Signal Period (ms)  *
; *                  ========================================================== *
; *                      48              12          1.05                       *
; *                       4              1           12.6                       *
; *                  ========================================================== *
; *                  * These timing could be incorrect when using code          *
; *                    optimisation or extended instruction mode,               *
; *                    or when having other interrupts enabled.                 *
; *                    Make sure to verify using the MPLAB SIM's Stopwatch      *
; *******************************************************************************
Remote@Wakeup
    movlb high(_USB#RAM_START)
    Btfss __USB_STAT,__REMOTEWAKEUP
    Return
    Bcf UCON,PP_SUSPND
    Bsf UCON,PP_RESUME
    Movlw 0X10
    Movwf PRODH
    Clrf PRODL
    Decfsz PRODL,F
    Bra $ - 2
    Decfsz PRODH,F
    Bra $ - 6
    Bcf UCON,PP_RESUME
    Return
; From usbdrv.c
;/******************************************************************************
; * Function:        void USB_SOF_Handler(void)                                *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        The USB host sends out a SOF packet to full-speed devices *
; *                  every 1 ms. This interrupt may be useful for isochronous  *
; *                  pipes. End designers should implement callback routine    *
; *                  as necessary.                                             *
; * Note:            None                                                      *
; ******************************************************************************
USB@SOF_Handler
; Callback routine here
;   Bcf UIR, SOFIF
;	  Return
; From usbdrv.c
; ********************************************************************************
; * Function:        void USBStallHandler(void)                                  *
; * PreCondition:    A STALL packet is sent to the host by the SIE.              *
; * Input:           None                                                        *
; * Output:          None                                                        *
; * Side Effects:    None                                                        *
; * Overview:        The STALLIF is set anytime the SIE sends out a STALL        *
; *                  packet regardless of which endpoint causes it.              *
; *                  A Setup transaction overrides the STALL function. A stalled *
; *                  endpoint stops stalling once it receives a setup packet.    *
; *                  In this case, the SIE will accepts the Setup packet and     *
; *                  set the TRNIF flag to notify the firmware. STALL function   *
; *                  for that particular endpoint pipe will be automatically     *
; *                  disabled (direction specific).                              *
; *                                                                              *
; *                  There are a few reasons for an endpoint to be stalled.      *
; *                  1. When a non-supported USB request is received.            *
; *                     Example: GET_DESCRIPTOR(DEVICE_QUALIFIER)                *
; *                  2. When an endpoint is currently halted.                    *
; *                  3. When the device class specifies that an endpoint must    *
; *                     stall in response to a specific event.                   *
; *                     Example: Mass Storage Device Class                       *
; *                              If the CBW is not valid, the device shall       *
; *                              STALL the Bulk-In pipe.                         *
; *                              See USB Mass Storage Class Bulk-only Transport  *
; *                              Specification for more details.                 *
; * Note:            UEPn.EPSTALL can be scanned to see which endpoint causes    *
; *                  the stall event.                                            *
; ********************************************************************************
Stall@Handler
;
; Does not really have to do anything here,
; even for the control endpoint.
; All BDs of Endpoint 0 are owned by SIE right now,
; but once a Setup Transaction is received, the ownership
; for EP0_OUT will be returned to CPU.
; When the Setup Transaction is serviced, the ownership
; for EP0_IN will then be forced back to CPU by firmware.
;
; NOTE: Above description is not quite true at this point.
;       It seems the SIE never returns the UOWN bit to CPU,
;       and a TRNIF is never generated upon successful
;       reception of a SETUP transaction.
;       Firmware work-around is implemented below.
;
NOLIST
if (UEP0 < 0X0F60)
LIST
    Movlb ((UEP0 >> 8) & 255)
NOLIST
endif
LIST
    Btfss UEP0,PP_EPSTALL
    Bra STL@EX
    movlb high(_USB#RAM_START)
    Movlw _USIE
    Cpfseq Ep0bO_Stat
    Bra STL@EX
    Movlw _USIE | _BSTALL
    Cpfseq Ep0bI_Stat
    Bra STL@EX
    Movlw _USIE | _DAT0 | __DTSEN | _BSTALL
    Movwf Ep0bO_Stat
NOLIST
if (UEP0 < 0X0F60)
LIST
    Movlb ((UEP0 >> 8) & 255)
NOLIST
endif
LIST
    Bcf UEP0,PP_EPSTALL
STL@EX
    Bcf UIR,PP_STALLIF
    Return
; From usbdrv.c
; ******************************************************************************
; * Function:        void Error@Handler(void)                                  *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        The purpose of this interrupt is mainly for debugging     *
; *                  during development. Check UEIR to see which error causes  *
; *                  the interrupt.                                            *
; * Note:            None                                                      *
; ******************************************************************************
Error@Handler
;
; Callback routine here
;
; UERRIF bit is a read-only bit and cannot be cleared directly.
; The correct way is to clear UEIR register.
;
;   clrf	UEIR			; (Ver 1.2)
;   return
; From usbdrv.c
; ******************************************************************************
; * Function:        void USBProtocolResetHandler(void)                        *
; * PreCondition:    A USB bus reset is received from the host.                *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    Currently, this routine flushes any pending USB           *
; *                  transactions. It empties out the USTAT FIFO. This action  *
; *                  might not be desirable in some applications.              *
; * Overview:        Once a USB bus reset is received from the host, this      *
; *                  routine should be called. It resets the device address to *
; *                  zero, disables all non-EP0 endpoints, initialises EP0 to  *
; *                  be ready for default communication, clears all USB        *
; *                  interrupt flags, unmasks applicable USB interrupts, and   *
; *                  re-initialises internal state-machine variables.          *
; * Note:            None                                                      *
; ******************************************************************************
Protocol@ResetHandler
NOLIST
if (UEIR < 0X0F60)
LIST
    Movlb ((UEIR >> 8) & 255)
NOLIST
endif
LIST
    Clrf UEIR
    Clrf UIR
NOLIST
if (UEIE < 0X0F60)
LIST
    Movlb ((UEIE >> 8) & 255)
NOLIST
endif
LIST
    Movlw 0X9F
    Movwf UEIE
    Movlw 0X7B
    Movwf UIE
    Clrf UADDR
    Lfsr 2,UEP1
NOLIST
#if (__Ep0bO == 512) ; Smaller USB devices only have 8 end points
LIST
    Movlw 7
NOLIST
#else
LIST
    Movlw 15
NOLIST
#endif
LIST
    Clrf POSTINC2
    Decfsz WREG,F
    Bra $ - 4
    Movlw __EP_CTRL | __HSHK_EN
    Movwf UEP0
    Bra $ + 10
    Bcf UIR,PP_TRNIF
    Bra $ + 2
    Bra $ + 2
    Bra $ + 2
    Btfsc UIR,PP_TRNIF
    Bra $ - 10
    Bcf UCON,PP_PKTDIS
    Rcall USBPrepareForNextSetupTrf
    movlb high(_USB#RAM_START)
    Bcf __USB_STAT,__REMOTEWAKEUP
    Clrf __USB_ACTIVE_CFG
    Movlw __DEFAULT_STATE
    Movwf __USB_DEVICE_STATE
    Return
; ******************************************************************************
; * Function:        void USBCheckCDCRequest(void)                             *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        Check the setup data packet to see if it                  *
; *                  knows how to handle a CDC request                         *
; * Note:            None                                                      *
; ******************************************************************************
Check@CDCRequest
    movlb high(_USB#RAM_START)
;
; If request recipient is not an interface then return
;
    Movf __SetupPkt,W
    Andlw 0x1F
    Sublw __RCPT_INTF
    Bnz Check@CDCRqstExit
;
; If request type is not class-specific then return
;
    Movf __SetupPkt,W
    Andlw 0x60
    Sublw (__CLASS << 5)
    Bnz Check@CDCRqstExit
;
; Interface ID must match interface numbers associated with
; CDC class, else return
;
    Movlw __CDC_COMM_INTF_ID
    Subwf SetupPkt_bIntfID,W
    Bz Check@CDCRqst1
    Movlw __CDC_DATA_INTF_ID
    Cpfseq SetupPkt_bIntfID
Check@CDCRqstExit
    Return
Check@CDCRqst1
    Movlw __SEND_ENCAPSULATED_COMMAND
    Cpfseq SetupPkt_bRequest
    Bra Check@CDCRqst2

    Movlw (__DUMMY_ENCAP_CMD_RESPONSE & 255)
    Movwf __Psrc
    Movlw ((__DUMMY_ENCAP_CMD_RESPONSE >> 8) & 255)
    Movwf __PsrcH

    Bcf __USB_STAT,__CTRL_TRF_MEM
    Movlw __DUMMY_LENGTH
    Movwf __wCount
    Clrf __wCountH
    Bra USBCDCSetSessionOwner
Check@CDCRqst2
    Movlw __GET_ENCAPSULATED_RESPONSE
    Cpfseq SetupPkt_bRequest
    Bra Check@CDCRqst3

    Movlw (__DUMMY_ENCAP_CMD_RESPONSE & 255)
    Movwf __Pdst
    Movlw ((__DUMMY_ENCAP_CMD_RESPONSE >> 8) & 255)
    Movwf __PdstH
    Bra USBCDCSetSessionOwner
Check@CDCRqst3
    Movlw __SET_COMM_FEATURE
    Cpfseq SetupPkt_bRequest
    Bra Check@CDCRqst4
    Return
Check@CDCRqst4
    Movlw __GET_COMM_FEATURE
    Cpfseq SetupPkt_bRequest
    Bra Check@CDCRqst5
    Return
Check@CDCRqst5
    Movlw __CLEAR_COMM_FEATURE
    Cpfseq SetupPkt_bRequest
    Bra Check@CDCRqst6
    Return
Check@CDCRqst6
    Movlw __SET_LINE_CODING
    Cpfseq SetupPkt_bRequest
    Bra Check@CDCRqst7

    Movlw (__Line_Coding & 255)
    Movwf __Pdst
    Movlw ((__Line_Coding >> 8) & 255)
    Movwf __PdstH

    Bra USBCDCSetSessionOwner
Check@CDCRqst7
    Movlw __GET_LINE_CODING
    Cpfseq SetupPkt_bRequest
    Bra Check@CDCRqst8
; Baud rate of the USB interface

    Movlw (115200 & 255)
    Movwf LineCoding_dwDTERate
    Movlw ((115200 >> 8) & 255)
    Movwf LineCoding_dwDTERateH
    Movlw ((115200 >> 16) & 255)
    Movwf LineCoding_dwDTERateHH
    Clrf LineCoding_dwDTERateHHH
    Clrf LineCoding_bCharFormat   ; 1 stop bit
    Clrf LineCoding_bParityType   ; No Parity
    Movlw 8
    Movwf LineCoding_bDataBits    ; 5, 6, 7, 8, or 16 data bits

    Movlw (__Line_Coding & 255)
    Movwf __Psrc
    Movlw ((__Line_Coding >> 8) & 255)
    Movwf __PsrcH

    Bcf __USB_STAT,__CTRL_TRF_MEM
    Movlw __LINE_CODING_LENGTH
    Movwf __wCount
    Clrf __wCountH
    Bra USBCDCSetSessionOwner
Check@CDCRqst8
    Movlw __SET_CONTROL_LINE_STATE
    Cpfseq SetupPkt_bRequest
    Bra Check@CDCRqst9
    Movff SetupPkt_wValue,__CONTROL_SIGNAL_BITMAP
    Bra USBCDCSetSessionOwner
Check@CDCRqst9
    Movlw __SEND_BREAK
    Cpfseq SetupPkt_bRequest
    Bra Check@CDCRqst10
    Return
Check@CDCRqst10
    Return
USBCDCSetSessionOwner
    Movlw __MUID_CDC
    Movwf __CTRL_TRF_SESSION_OWNER
    Return
; ******************************************************************************
; * Function:        void CDCInitEP(void)                                      *
; * PreCondition:    None                                                      *
; * Input:           None                                                      *
; * Output:          None                                                      *
; * Side Effects:    None                                                      *
; * Overview:        Initialise CDC endpoints, buffer descriptors,             *
; *                  internal state-machine, and variables.                    *
; *                  It should be called after the USB host has sent out a     *
; *                  SET_CONFIGURATION request.                                *
; *                  See USBStdSetCfgHandler() in usb9.c for examples.         *
; * Note:            None                                                      *
; ******************************************************************************
CDCInit@EP
    Movlw 2
    Movwf FSR0L
    Lfsr 1,__CDC_NOTICE
    Movlw __CDC_INT_EP_SIZE
; Initialise the endpoint
    Rcall InitEP@In
    Movlw 3
    Movwf FSR0L
    Lfsr 1,__CDC_DATA_RX
    Movlw __CDC_BULK_OUT_EP_SIZE
; Initialise the endpoint
    Movwf POSTINC1
    Lfsr 2,UEP0
    Movf FSR0L,W
    Addwf FSR2L,F
    Movlw __EP_OUT | __HSHK_EN
    Iorwf INDF2,F
    Rlncf FSR0L,W
    Rlncf WREG,F
    Rlncf  WREG,F
    Lfsr 0,__Ep0bO
    Addwf FSR0L,F
    Movlw __ADRL
    Movff FSR1L,PLUSW0
    Movlw __ADRH
    Movff FSR1H,PLUSW0
    Movf POSTDEC1,W
    Incf FSR0L,F
    Movff INDF1,POSTDEC0
    Movlw _USIE | _DAT0 | _DTSEN
    Movwf INDF0

    Movlw 3
    Movwf FSR0L
    Lfsr 1,__CDC_DATA_TX
    Movlw __CDC_BULK_IN_EP_SIZE
; Initialise the endpoint
InitEP@In
    Movwf POSTINC1
    Lfsr 2,UEP0
    Movf FSR0L,W
    Addwf FSR2L,F
    Movlw __EP_IN | __HSHK_EN
    Iorwf INDF2,F
    Rlncf FSR0L,W
    Rlncf WREG,F
    Rlncf WREG,F
    Lfsr 0,__EP0BI
    Addwf FSR0L,F
    Movlw __ADRL
    Movff FSR1L,PLUSW0
    Movlw __ADRH
    Movff FSR1H,PLUSW0
    Movlw _UCPU | _DAT1
    Movwf INDF0
    Return
